@@ -16,22 +16,15 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
  *
  */
- 
- 
-#ifndef _CALLBACKS_H_
-#define _CALLBACKS_H_
 
-#if 0
-#include <gnome.h>
-#include <gal/e-table/e-table.h>
-#endif
+
+#ifndef _PROCMAN_CALLBACKS_H_
+#define _PROCMAN_CALLBACKS_H_
+
 #include "procman.h"
 
 
-
-void
-cb_properties_activate                (GtkMenuItem     *menuitem,
-                                        gpointer         user_data);
+void		cb_properties_activate (GtkMenuItem *menuitem, gpointer user_data);
 
 void		cb_show_memory_maps (GtkMenuItem *menuitem, gpointer data);
 void		cb_renice (GtkMenuItem *menuitem, gpointer data);
@@ -44,27 +37,22 @@ void		cb_preferences_activate (GtkMenuItem *menuitem, gpointer user_data);
 
 void		cb_about_activate (GtkMenuItem *menuitem, gpointer user_data);
 
-void		cb_app_exit (GtkObject *object, gpointer user_data); 
+void		cb_app_exit (GtkObject *object, gpointer user_data);
 gboolean	cb_app_delete (GtkWidget *window, GdkEventAny *event, gpointer data);
 #if 0
 gboolean	cb_close_simple_dialog (GnomeDialog *dialog, gpointer data);
-#endif                                        
-void		cb_all_process_menu_clicked (GtkWidget *widget, gpointer data);	
-					 
-void		cb_my_process_menu_clicked (GtkWidget *widget, gpointer data);  
-					 
-void		cb_running_process_menu_clicked	(GtkWidget *widget, gpointer data);
+#endif
+void		cb_proc_combo_changed (GtkComboBox *combo, gpointer data);
 
-void		cb_favorites_menu_clicked (GtkWidget *widget, gpointer data);
 void		cb_end_process_button_pressed (GtkButton *button, gpointer data);
 void		cb_logout (GtkButton *button, gpointer data);
 
 void		popup_menu_renice (GtkMenuItem *menuitem, gpointer data);
 void		popup_menu_show_memory_maps (GtkMenuItem *menuitem, gpointer data);
 void		popup_menu_hide_process (GtkMenuItem *menuitem, gpointer data);
-void 		popup_menu_end_process (GtkMenuItem *menuitem, gpointer data);
-void 		popup_menu_kill_process (GtkMenuItem *menuitem, gpointer data);
-void 		popup_menu_about_process (GtkMenuItem *menuitem, gpointer data);
+void		popup_menu_end_process (GtkMenuItem *menuitem, gpointer data);
+void		popup_menu_kill_process (GtkMenuItem *menuitem, gpointer data);
+void		popup_menu_about_process (GtkMenuItem *menuitem, gpointer data);
 
 void		cb_info_button_pressed (GtkButton *button, gpointer user_data);
 void		cb_search (GtkEditable *editable, gpointer data);
@@ -76,18 +64,24 @@ void		cb_mem_color_changed (GnomeColorPicker *cp, guint r, guint g, guint b,
 				      guint a, gpointer data);
 void		cb_swap_color_changed (GnomeColorPicker *cp, guint r, guint g, guint b,
 				      guint a, gpointer data);
+void		cb_bg_color_changed (GnomeColorPicker *cp, guint r, guint g, guint b,
+				      guint a, gpointer data);
+void		cb_frame_color_changed (GnomeColorPicker *cp, guint r, guint g, guint b,
+				      guint a, gpointer data);
+
 
 void		cb_row_selected (GtkTreeSelection *selection, gpointer data);
-gboolean	cb_tree_button_pressed (GtkWidget *widget, GdkEventButton *event, 
+void		cb_tree_row_activated (GtkTreeView *view, GtkTreePath *path,
+				       GtkTreeViewColumn *column, gpointer data);
+gboolean	cb_tree_popup_menu (GtkWidget *widget, gpointer data);
+gboolean	cb_tree_button_pressed (GtkWidget *widget, GdkEventButton *event,
 					gpointer data);
-gint		cb_tree_key_press (GtkWidget *widget, GdkEventKey *event, gpointer data);
 
-					 
-void		cb_switch_page (GtkNotebook *nb, GtkNotebookPage *page, 
-			        gint num, gpointer data);
+void		cb_switch_page (GtkNotebook *nb, GtkNotebookPage *page,
+				gint num, gpointer data);
 
-gint 		cb_update_disks (gpointer data);
+gint		cb_update_disks (gpointer data);
 gint		cb_timeout (gpointer data);
 
 
-#endif
+#endif /* _PROCMAN_CALLBACKS_H_ */
